//-----------------------------------------------------------------------------
// File: EnhancedMesh.cpp
//
// Desc: Sample showing enhanced meshes in D3D
//
// Copyright (c) 1999-2000 Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------
(*----------------------------------------------------------------------------*
 *  Direct3D sample from DirectX 8.0 SDK Delphi adaptation by Alexey Barkovoy *
 *  E-Mail: clootie@reactor.ru                                                *
 *                                                                            *
 *  Modified: 14-Jul-2001                                                     *
 *                                                                            *
 *  Lastest version can be downloaded from:                                   *
 *     http://groups.yahoo.com/group/JEDI-DirectXExamples                     *
 *     http://clootie.narod.ru                                                *
 *----------------------------------------------------------------------------*)

unit EnhMCode;

{$A+}  // Word alignment data
{$OA+} // Objects and structures align
{$Z4}  // Set minimum size of enumerated type to 4

interface

uses
  Windows, Messages, Math, CommDlg, D3D8, D3DX8, D3DApp, D3DFont, D3DUtil, DXUtil;

const
  IDM_OPENFILE          = 40011;

type
  PD3DXMaterialsArray = ^TD3DXMaterialsArray;
  TD3DXMaterialsArray = array [0..0] of TD3DXMaterial;

  TDirect3DTexturesArray = ^TTDirect3DTexturesArray;
  TTDirect3DTexturesArray = array [0..0] of IDirect3DTexture8;

  //-----------------------------------------------------------------------------
  // Name: class CMyD3DApplication
  // Desc: Application class. The base class (CD3DApplication) provides the
  //       generic functionality needed in all Direct3D samples. CMyD3DApplication
  //       adds functionality specific to this sample program.
  //-----------------------------------------------------------------------------
  CMyD3DApplication = ^TMyD3DApplication;
  TMyD3DApplication = object(TD3DApplication)
  private
    m_strMeshFilename:  array [0..511] of Char;
    m_strInitialDir:    array [0..511] of Char;

    m_pMeshSysMem:      ID3DXMesh;          // system memory version of mesh, lives through resize's
    m_pMeshEnhanced:    ID3DXMesh;          // vid mem version of mesh that is enhanced
    m_dwNumSegs:        Cardinal;           // number of segments per edge (tesselation level)
    m_pMaterials:       PD3DXMaterialsArray;// pointer to material info in m_pbufMaterials
    m_ppTextures:       TDirect3DTexturesArray;// array of textures, entries are NULL if no texture specified
    m_dwNumMaterials:   DWORD;              // number of materials

    m_pFont:            CD3DFont;
    m_ArcBall:          CD3DArcBall;        // mouse rotation utility
    m_vObjectCenter:    TD3DXVector3;       // Center of bounding sphere of object
    m_fObjectRadius:    Single;             // Radius of bounding sphere of object

    m_pbufMaterials:    ID3DXBuffer;        // contains both the materials data and the filename strings
    m_pbufAdjacency:    ID3DXBuffer;        // Contains the adjacency info loaded with the mesh

    m_bUseHWNPatches:   BOOL;

    function GenerateEnhancedMesh(cNewNumSegs: Cardinal): HResult;

    function OneTimeSceneInit: HResult; virtual;
    function InitDeviceObjects: HResult; virtual;
    function RestoreDeviceObjects: HResult; virtual;
    function InvalidateDeviceObjects: HResult; virtual;
    function DeleteDeviceObjects: HResult; virtual;
    function Render: HResult; virtual;
    function FrameMove: HResult; virtual;
    function FinalCleanup: HResult; virtual;
  public
    constructor Create;
    destructor Destroy; virtual;
    function MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT; virtual;
 end;

implementation

uses
  Strings;

{ CMyD3DApplication }

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
constructor TMyD3DApplication.Create;
begin
  inherited Create(0);

  m_strWindowTitle    := 'Enhanced Mesh - N-Patches';
  m_bUseDepthBuffer   := TRUE;
  m_bShowCursorWhenFullscreen := TRUE;

  m_pMeshSysMem     := nil;
  m_pMeshEnhanced   := nil;
  m_pMaterials      := nil;
  m_ppTextures      := nil;
  m_dwNumMaterials  := 0;
  m_pbufMaterials   := nil;
  m_pbufAdjacency   := nil;
  m_dwNumSegs       := 2;
  m_bUseHWNPatches  := FALSE;

  New(m_pFont,    Create('Arial', 12, D3DFONT_BOLD));

  StrCopy(m_strInitialDir, DXUtil_GetDXSDKMediaPath);
  m_strMeshFilename:= 'tiger.x'+#0;

  New(m_ArcBall, Create);
end;

destructor TMyD3DApplication.Destroy;
begin
  m_ArcBall^.Free;
  inherited Destroy;
end;

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function TMyD3DApplication.OneTimeSceneInit: HResult;
begin
  // Set cursor to indicate that user can move the object with the mouse
{$IFDEF WIN64}
  SetClassLongPtr(m_hWnd, GCLP_HCURSOR, LoadCursor(0, IDC_SIZEALL));
{$ELSE}
  SetClassLong(m_hWnd, GCL_HCURSOR, LoadCursor(0, IDC_SIZEALL));
{$ENDIF}
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function TMyD3DApplication.FrameMove: HResult;
var
  matWorld: TD3DXMatrix;
  matRotationInverse: TD3DXMatrix;
  matView: TD3DXMatrix;
begin
   // Setup world matrix
  D3DXMatrixTranslation(matWorld, -m_vObjectCenter.x,
                                  -m_vObjectCenter.y,
                                  -m_vObjectCenter.z);
  D3DXMatrixInverse(matRotationInverse, nil, m_ArcBall^.GetRotationMatrix^);
  D3DXMatrixMultiply(matWorld, matWorld, matRotationInverse );
  D3DXMatrixMultiply(matWorld, matWorld, m_ArcBall^.GetTranslationMatrix^);
  m_pd3dDevice.SetTransform(D3DTS_WORLD, matWorld);

  D3DXMatrixLookAtLH(matView, D3DXVector3(0, 0, -2*m_fObjectRadius),
                              D3DXVector3(0, 0, 0),
                              D3DXVector3(0, 1, 0));
  m_pd3dDevice.SetTransform(D3DTS_VIEW,  matView);

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function TMyD3DApplication.Render: HResult;
var
  fNumSegs: Single;
  i: Integer;

  Str1, Str2: String;
  pBuff1, pBuff2: array[0..100] of Char;
begin
  // Clear the backbuffer
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER, $00FF, 1, 0);

  // Begin the scene
  if SUCCEEDED(m_pd3dDevice.BeginScene) then
  begin
    if (m_bUseHWNPatches) then
    begin
      fNumSegs:= m_dwNumSegs;
      m_pd3dDevice.SetRenderState(D3DRS_PATCHSEGMENTS, PDWord(@fNumSegs)^);
    end;

    // set and draw each of the materials in the mesh
    for i:= 0 to (m_dwNumMaterials - 1) do
    begin
      m_pd3dDevice.SetMaterial(m_pMaterials^[i].MatD3D);
      m_pd3dDevice.SetTexture(0, m_ppTextures^[i]);

      m_pMeshEnhanced.DrawSubset(i);
    end;

    if m_bUseHWNPatches then
    begin
      m_pd3dDevice.SetRenderState(D3DRS_PATCHSEGMENTS, 0);
    end;

    // Output stats
    begin
      // Output statistics
      m_pFont^.DrawText(   2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats, 0);
      m_pFont^.DrawText(   2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats, 0);

      pBuff1:= 'NumSegs:';
      Str2:= IntToStr(m_dwNumSegs) + ' ';
      StrPCopy(pBuff2, Str2);
      m_pFont^.DrawText(   2, 40, $ffffff00, pBuff1, 0);
      m_pFont^.DrawText( 150, 40, $ffffffff, pBuff2, 0);
      pBuff1:= 'NumFaces:';
      Str2:= IntToStr(m_pMeshEnhanced.GetNumFaces) + ' ';
      StrPCopy(pBuff2, Str2);
      m_pFont^.DrawText(   2, 60, $ffffff00, pBuff1, 0);
      m_pFont^.DrawText( 150, 60, $ffffffff, pBuff2, 0);
      Str1:= 'NumVertices:';
      Str2:= IntToStr(m_pMeshEnhanced.GetNumVertices) + ' ';
      StrPCopy(pBuff2, Str2);
      m_pFont^.DrawText(   2, 80, $ffffff00, pBuff1, 0);
      m_pFont^.DrawText( 150, 80, $ffffffff, pBuff2, 0);
    end;

    // End the scene.
    m_pd3dDevice.EndScene;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: This creates all device-dependent managed objects, such as managed
//       textures and managed vertex buffers.
//-----------------------------------------------------------------------------
function TMyD3DApplication.InitDeviceObjects: HResult;
var
  pVB: IDirect3DVertexBuffer8;
  pVertices: PByte;
  pTempMesh: ID3DXMesh;
  strMeshPath: array[0..511] of Char;
  strTexturePath: array[0..511] of Char;
  i: Integer;
begin
  pVertices := nil;

  // Initialize the font's internal textures
  m_pFont^.InitDeviceObjects(m_pd3dDevice);

  // Load the mesh from the specified file
  DXUtil_FindMediaFile(strMeshPath, m_strMeshFilename);

  Result:= D3DXLoadMeshFromX(strMeshPath, D3DXMESH_SYSTEMMEM, m_pd3dDevice,
                          @m_pbufAdjacency, @m_pbufMaterials, @m_dwNumMaterials,
                          m_pMeshSysMem);
  if FAILED(Result) then
  begin
    Result := D3DAPPERR_MEDIANOTFOUND;
    Exit;
  end;

  // Lock the vertex buffer, to generate a simple bounding sphere
  Result:= m_pMeshSysMem.GetVertexBuffer(pVB);
  if FAILED(Result) then Exit;

  Result:= pVB.Lock(0, 0, pVertices, 0);
  if FAILED(Result) then Exit; // VB will be released by Delphi compiler

  Result:= D3DXComputeBoundingSphere(pVertices^, m_pMeshSysMem.GetNumVertices,
                                  m_pMeshSysMem.GetFVF, m_vObjectCenter,
                                  m_fObjectRadius);
  if FAILED(Result) then
  begin
    pVB.Unlock;
    pVB:= nil;
    Exit;
  end;

  if (0 = m_dwNumMaterials) then
  begin
    pVB.Unlock;
    pVB:= nil;
    Result := E_INVALIDARG;
    Exit;
  end;

  // Get the array of materials out of the Result :=ed buffer, allocate a
  // texture array, and load the textures
  m_pMaterials := PD3DXMaterialsArray(m_pbufMaterials.GetBufferPointer);
  GetMem(m_ppTextures, SizeOf(IDirect3DTexture8)*m_dwNumMaterials);

  for i:= 0 to (m_dwNumMaterials - 1) do
  begin
    DXUtil_FindMediaFile(strTexturePath, m_pMaterials^[i].pTextureFilename);
    if FAILED(D3DXCreateTextureFromFile(m_pd3dDevice, strTexturePath, m_ppTextures^[i]))
    then m_ppTextures^[i]:= nil;
  end;

  pVB.Unlock;
  pVB:= nil;

  // Make sure there are normals, which are required for the tesselation
  // enhancement
  if (m_pMeshSysMem.GetFVF and D3DFVF_NORMAL) = 0 then
  begin
    Result:= m_pMeshSysMem.CloneMeshFVF(m_pMeshSysMem.GetOptions,
                                      m_pMeshSysMem.GetFVF or D3DFVF_NORMAL,
                                      m_pd3dDevice, pTempMesh);
    if FAILED(Result) then Exit;

    D3DXComputeNormals(pTempMesh);

    m_pMeshSysMem:= nil;
    m_pMeshSysMem:= pTempMesh;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: GenerateEnhancedMesh()
// Desc:
//-----------------------------------------------------------------------------
function TMyD3DApplication.GenerateEnhancedMesh(cNewNumSegs: Cardinal): HResult;
var
  pMeshEnhancedSysMem: ID3DXMesh;
  pMeshTemp: ID3DXMesh;
  c: Single;
begin
  // if using hw, just copy the mesh
  if (m_bUseHWNPatches) then
  begin
    Result:= m_pMeshSysMem.CloneMeshFVF(D3DXMESH_WRITEONLY or D3DXMESH_NPATCHES,
                            m_pMeshSysMem.GetFVF, m_pd3dDevice, pMeshTemp);
    if FAILED(Result) then Exit;
  end else  // tesselate the mesh in sw
  begin
    c:= cNewNumSegs;
    // Create an enhanced version of the mesh, will be in sysmem since source is
    Result:= D3DXTesselateMesh(m_pMeshSysMem, PDWord(m_pbufAdjacency.GetBufferPointer),
                            c, FALSE, pMeshEnhancedSysMem);
    if FAILED(Result) then Exit;

    // Make a vid mem version of the mesh
    Result:= pMeshEnhancedSysMem.CloneMeshFVF(D3DXMESH_WRITEONLY, m_pMeshSysMem.GetFVF,
               m_pd3dDevice, pMeshTemp);
    if FAILED(Result) then Exit; // pMeshEnhancedSysMem:= nil - automatically

    // Latch in the enhanced mesh
    pMeshEnhancedSysMem:= nil;
  end;

  m_pMeshEnhanced := nil;
  m_pMeshEnhanced := pMeshTemp;
  m_dwNumSegs     := cNewNumSegs;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Restore device-memory objects and state after a device is created or
//       resized.
//-----------------------------------------------------------------------------
function TMyD3DApplication.RestoreDeviceObjects: HResult;
var
  light: TD3DLight8;
  matProj: TD3DXMatrix;
  fAspect: Single;
begin
  m_bUseHWNPatches:= (m_d3dCaps.DevCaps and D3DDEVCAPS_NPATCHES) <> 0;

  Result:= GenerateEnhancedMesh(m_dwNumSegs);
  if FAILED(Result) then Exit;

  // Setup render state
  m_pd3dDevice.SetRenderState(D3DRS_LIGHTING,     iTRUE);
  m_pd3dDevice.SetRenderState(D3DRS_DITHERENABLE, iTRUE);
  m_pd3dDevice.SetRenderState(D3DRS_ZENABLE,      iTRUE);
  m_pd3dDevice.SetRenderState(D3DRS_AMBIENT,      $33333333);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_LINEAR);

  // Setup the light
  D3DUtil_InitLight(light, D3DLIGHT_DIRECTIONAL, 0.0,-1.0, 1.0);
  m_pd3dDevice.SetLight(0, light);
  m_pd3dDevice.LightEnable(0, TRUE);

  // Setup the arcball parameters
  m_ArcBall^.SetWindow(m_d3dsdBackBuffer.Width, m_d3dsdBackBuffer.Height, 0.85);

  // Setup the projection matrix
  fAspect := m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;
  D3DXMatrixPerspectiveFovLH(matProj, D3DX_PI/4, fAspect,
                             m_fObjectRadius/64.0, m_fObjectRadius*200.0);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, matProj);

  // Restore the font
  m_pFont^.RestoreDeviceObjects;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc: Called when the device-dependent objects are about to be lost.
//-----------------------------------------------------------------------------
function TMyD3DApplication.InvalidateDeviceObjects: HResult;
begin
  m_pFont^.InvalidateDeviceObjects;

  m_pMeshEnhanced:= nil;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function TMyD3DApplication.DeleteDeviceObjects: HResult;
var
  i: Integer;
begin
  m_pFont^.DeleteDeviceObjects;

  for i:= 0 to m_dwNumMaterials - 1 do SAFE_RELEASE(m_ppTextures^[i]);
  FreeMem(m_ppTextures, SizeOf(IDirect3DTexture8)*m_dwNumMaterials);
  m_ppTextures:= nil;
  SAFE_RELEASE(m_pMeshSysMem);
  SAFE_RELEASE(m_pbufMaterials);
  m_dwNumMaterials := 0;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function TMyD3DApplication.FinalCleanup: HResult;
begin
  m_pFont^.Free;
  m_pFont:= nil;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Message proc function to handle key and menu input
//-----------------------------------------------------------------------------
function TMyD3DApplication.MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM;
  lParam: LPARAM): LRESULT;
const
  ofn: TOpenFilename = (
    lStructSize: SizeOf(TOpenFilename);
    lpstrFilter: '.X Files (.x)'#0'*.x'#0#0; nFilterIndex: 1;
    nMaxFile: 512;
    nMaxFileTitle: 512;
    lpstrTitle: 'Open Mesh File';
    Flags: OFN_FILEMUSTEXIST;
    nFileExtension: 1
  );
var
  g_strFilename: array [0..511] of Char;
  pLastSlash: PChar;
begin
  // Pass mouse messages to the ArcBall so it can build internal matrices
  m_ArcBall^.HandleMouseMessages(hWnd, uMsg, wParam, lParam);

  // Trap context menu
  if (WM_CONTEXTMENU = uMsg) then
  begin
    Result := 0;
    Exit;
  end;

  // Handle key presses
  if (WM_KEYDOWN = uMsg) then
  begin
    if (VK_UP = wParam)
    then GenerateEnhancedMesh(m_dwNumSegs + 1);

    if (VK_DOWN = wParam)
    then GenerateEnhancedMesh(Max(1, m_dwNumSegs - 1));

    if (Ord('W') = wParam)
    then m_pd3dDevice.SetRenderState(D3DRS_FILLMODE, D3DFILL_WIREFRAME);

    if (Ord('S') = wParam)
    then m_pd3dDevice.SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);
  end
  else if (uMsg = WM_COMMAND) then
  begin
    // Handle the open file command
    if (LOWORD(wParam) = IDM_OPENFILE) then
    begin
      g_strFilename:= '';

      // Display the OpenFileName dialog. Then, try to load the specified file
      ofn.lpstrFile:= m_strMeshFilename;
      ofn.lpstrFileTitle:= g_strFilename;
      ofn.lpstrInitialDir:= m_strInitialDir;

      if GetOpenFileName(ofn) then
      begin
        StrCopy(m_strInitialDir, m_strMeshFilename);
        pLastSlash:= StrPos(m_strInitialDir, '\');
        if (pLastSlash <> nil) then pLastSlash^ := #0;

        SetCurrentDirectory(m_strInitialDir);

        // Destroy and recreate everything
        InvalidateDeviceObjects;
        DeleteDeviceObjects;
        InitDeviceObjects;
        RestoreDeviceObjects;
      end;
    end;
  end;

  // Pass remaining messages to default handler
  Result := inherited MsgProc(hWnd, uMsg, wParam, lParam);
end;

end.
